
from __future__ import print_function

import os

here = os.path.abspath(os.path.split(__file__)[0])


def get_doc(lines, indent='    '):
    doc = None

    # find start doc
    for i, line in enumerate(lines):
        # search for end of code
        if len(line) - len(line.lstrip(' ')) < len(indent):
            break

        if line.startswith(indent + '"""'):
            lines_to_be_parsed = lines[i+1:]
            doc = line
            break

    if doc is None or len(doc.split('"""')) > 2:
        return doc

    # find end doc
    for i, line in enumerate(lines_to_be_parsed):
        doc += '\n' + line
        if '"""' in line:
            break

    return doc

strings_to_be_deleted = [
    ' DTYPEf_t[:, :, ::1] ', ' DTYPEc_t[:, :, ::1] ',
    ' DTYPEf_t[:, ::1] ', ' DTYPEc_t[:, ::1] ', ' int ']


def get_function_code(lines):

    signature = lines[0].replace(' cpdef ', ' def ')

    if not signature.endswith('):'):
        for i, line in enumerate(lines[1:]):
            signature += ' ' + line.strip()
            if line.endswith('):'):
                break
    else:
        i = 0

    lines_to_be_parsed = lines[i+1:]

    for s in strings_to_be_deleted:
        signature = signature.replace(s, ' ')

    func_lines = [signature]

    indent = ' ' * 8

    doc = get_doc(lines_to_be_parsed, indent=indent)

    if doc is None:
        doc = indent + 'pass'

    func_lines.append(doc)

    return '\n'.join(func_lines)


def create_fake_mod(dimension):

    with open(os.path.join(
            here, 'template{dim}d_mako.pyx'.format(dim=dimension)), 'r') as f:
        lines_text = f.read().splitlines()

    # find start class
    for i, line in enumerate(lines_text):
        if 'cdef class ' in line:
            lines_class = lines_text[i+1:]
            break

    # get docstring of the class
    doc_class = get_doc(lines_class)

    lines_function_class = lines_class[1:]

    # find functions
    functions_codes = []
    for i, line in enumerate(lines_function_class):
        if line.startswith('    def ') or line.startswith('    cpdef '):
            if ' __dealloc__(' in line or '__cinit__' in line:
                continue
            functions_codes.append(get_function_code(lines_function_class[i:]))

    code = (
        '\nclass FFT{dim}dFakeForDoc(object):\n'.format(dim=dimension) +
        doc_class + '\n' + '\n\n'.join(functions_codes) + '\n')

    name = '../fluidfft/fft{dim}d/fake_mod_fft{dim}d_for_doc.py'.format(
        dim=dimension)

    with open(os.path.join(here, name), 'w') as f:
        f.write(code)

for dim in range(2, 4):
    create_fake_mod(dim)



#cython: embedsignature=True

# DEF MPI4PY = 0

cimport cython

cimport numpy as np
import numpy as np
np.import_array()

try:
    from mpi4py import MPI
except ImportError:
    nb_proc = 1
    rank = 0
else:
    comm = MPI.COMM_WORLD
    nb_proc = comm.size
    rank = comm.Get_rank()

# IF MPI4PY:
from mpi4py cimport MPI

#from mpi4py.mpi_c cimport *
from mpi4py.libmpi cimport *

    # fix a bug arising when using a recent version of mpi4py
cdef extern from 'mpi-compat.h': pass

# we define python and c types for physical and Fourier spaces
DTYPEb = np.uint8
ctypedef np.uint8_t DTYPEb_t
DTYPEi = np.int
ctypedef np.int_t DTYPEi_t
DTYPEf = np.float64
ctypedef np.float64_t DTYPEf_t
DTYPEc = np.complex128
ctypedef np.complex128_t DTYPEc_t




import os
from os.path import join
from datetime import datetime

from mako.template import Template

here = os.path.abspath(os.path.dirname(__file__))


def modification_date(filename):
    t = os.path.getmtime(filename)
    return datetime.fromtimestamp(t)


template2d_pyx = Template(filename=join(here, 'template2d_mako.pyx'))
template2d_pxd = Template(filename=join(here, 'template2d_mako.pxd'))

template3d_pyx = Template(filename=join(here, 'template3d_mako.pyx'))
template3d_pxd = Template(filename=join(here, 'template3d_mako.pxd'))


def make_file(module_name, class_name):

    name_pyx = join(here, module_name + '_cy.pyx')
    name_pxd = join(here, module_name + '_pxd.pxd')

    if module_name.startswith('fft2d'):
        t_pyx = template2d_pyx
        t_pxd = template2d_pxd
    elif module_name.startswith('fft3d'):
        t_pyx = template3d_pyx
        t_pxd = template3d_pxd

    if not os.path.exists(name_pyx):
        hastomake = True
    else:
        if modification_date(name_pyx) < modification_date(t_pyx.filename):
            hastomake = True
        else:
            hastomake = False

    if hastomake:
        with open(name_pyx, 'w') as f:
            f.write(t_pyx.render(
                module_name=module_name, class_name=class_name))

    if not os.path.exists(name_pxd):
        hastomake = True
    else:
        if modification_date(name_pxd) < modification_date(t_pxd.filename):
            hastomake = True
        else:
            hastomake = False

    if hastomake:
        with open(name_pxd, 'w') as f:
            f.write(t_pxd.render(
                module_name=module_name, class_name=class_name))


def make_pyx_files():
    variables = (
        ('fft2d_with_fftw1d', 'FFT2DWithFFTW1D'),
        ('fft2d_with_fftw2d', 'FFT2DWithFFTW2D'),
        ('fft2d_with_cufft', 'FFT2DWithCUFFT'),
        ('fft2dmpi_with_fftw1d', 'FFT2DMPIWithFFTW1D'),
        ('fft2dmpi_with_fftwmpi2d', 'FFT2DMPIWithFFTWMPI2D'),
        ('fft3d_with_fftw3d', 'FFT3DWithFFTW3D'),
        ('fft3dmpi_with_fftwmpi3d', 'FFT3DMPIWithFFTWMPI3D'),
        ('fft3dmpi_with_pfft', 'FFT3DMPIWithPFFT'),
        ('fft3dmpi_with_p3dfft', 'FFT3DMPIWithP3DFFT'),
        ('fft3d_with_cufft', 'FFT3DWithCUFFT'),
    )

    for module_name, class_name in variables:
        make_file(module_name, class_name)


if __name__ == '__main__':
    make_pyx_files()




from __future__ import print_function, division

import unittest

from fluiddyn.util import mpi

from fluidfft.fft3d import get_classes_seq, get_classes_mpi

n = 16

rank = mpi.rank
nb_proc = mpi.nb_proc


def make_test_function(cls, sequencial):

    def test(self):
        if sequencial and rank > 0:
            return
        o = cls(n, n)
        o.run_tests()

    return test


class Tests3D(unittest.TestCase):
    pass


for name, cls in get_classes_seq().items():
    if cls is None:
        continue
    setattr(Tests3D, 'test_{0}'.format(name), make_test_function(cls, True))

for name, cls in get_classes_mpi().items():
    if cls is None:
        continue
    setattr(Tests3D, 'test_{0}'.format(name), make_test_function(cls, False))


if __name__ == '__main__':
    unittest.main()



"""Maker for fluidfft
=====================

The compilation process of the normal build_ext is not adapted for this project
since many files are common to different extensions but can be computed only
once.

All files are compiled manually and we launch the compilations asynchronously.

Notes
-----

- compilation:

  $CC $OPT $BASECFLAGS $CONFIGURE_CFLAGS $CCSHARED \
    $(sys.prefix)/include/python$('.'.join(sys.version_info[:2])) \
    -c path/source.cpp -o build/temp.linux-x86_64-2.7/path/source.o

- linker:

  $CXX -shared \
    -Wl,-rpath=$(sys.prefix)/lib/python$('.'.join(sys.version_info[:2])) \
    $(list object files) -L/opt/python/2.7.9/lib -lpython2.7 \
    -o build/temp.linux-x86_64-2.7/path/to/mod.so

  $CXX -shared ~$(LDSHARED) $(CONFIGURE_LDFLAGS) $(CONFIGURE_CFLAGS) \
  $(list object files)

"""

from __future__ import print_function

import sys
import platform
from time import sleep
import os
from datetime import datetime
from distutils import sysconfig
import subprocess
from copy import copy

config_vars = sysconfig.get_config_vars()

try:
    from Cython.Compiler.Main import \
        CompilationOptions, \
        default_options as cython_default_options, \
        compile_single as cython_compile
    can_import_cython = True
except ImportError:
    can_import_cython = False


short_version = '.'.join([str(i) for i in sys.version_info[:2]])

path_lib_python = os.path.join(sys.prefix, 'lib', 'python' + short_version)
# path_include_python = os.path.join(
#     sys.prefix, 'include', 'python' + short_version)

path_tmp = 'build/temp.' + '-'.join(
    [platform.system().lower(), platform.machine(), short_version])


def modification_date(filename):
    t = os.path.getmtime(filename)
    return datetime.fromtimestamp(t)


class CompilationError(Exception):
    def __init__(self, out, err):
        self.message = '\nstdout:\n {}\nstderr:\n{}'.format(out, err)
        super(CompilationError, self). __init__()

    def __str__(self):
        return super(CompilationError, self).__str__() + self.message


class Extension(object):
    def __init__(self, name, sources=None, language=None):
        self.name = name
        self.sources = sources
        self.language = language


def make_cpp_from_pyx(cpp_file, pyx_file, full_module_name=None, options=None):
    path_dir = os.path.split(cpp_file)[0]
    if not os.path.exists(path_dir):
        os.makedirs(path_dir)
    if not os.path.exists(cpp_file) or \
       modification_date(cpp_file) < modification_date(pyx_file):

        if not can_import_cython:
            raise ImportError('Can not import Cython.')

        include_path = None
        if options is not None and 'include_dirs' in options.keys():
            include_path = list(options['include_dirs'])

        options = CompilationOptions(
            cython_default_options,
            include_path=include_path,
            output_file=cpp_file,
            cplus=True,
            compile_time_env=None)
        result = cython_compile(pyx_file, options=options,
                                full_module_name=full_module_name)

        print('cythonize ' + pyx_file)

        return result


def make_obj_from_cpp(obj_file, cpp_file, options=None):
    path_dir = os.path.split(obj_file)[0]
    if not os.path.exists(path_dir):
        os.makedirs(path_dir)
    if not os.path.exists(obj_file) or \
       modification_date(obj_file) < modification_date(cpp_file):

        keys = ['CC', 'OPT', 'BASECFLAGS', 'CFLAGS', 'CCSHARED']

        conf_vars = copy(config_vars)

        if options is not None:
            for k, v in options.items():
                conf_vars[k] = v

        command = ' '.join([conf_vars[k] for k in keys])

        if cpp_file.endswith('.cu'):
            command = (
                'nvcc -m64 '
                '-gencode arch=compute_20,code=sm_20 '
                '-gencode arch=compute_30,code=sm_30 '
                '-gencode arch=compute_32,code=sm_32 '
                '-gencode arch=compute_35,code=sm_35 '
                '-gencode arch=compute_50,code=sm_50 '
                '-gencode arch=compute_50,code=compute_50 -Xcompiler -fPIC')

        command = [w for w in command.split()
                   if w not in ['-g', '-DNDEBUG', '-Wstrict-prototypes']]

        include_dirs = [conf_vars['INCLUDEPY']]

        if 'cufft' in cpp_file:
            include_dirs.extend([
                '/opt/cuda/NVIDIA_CUDA-6.0_Samples/common/inc/'
                 ])

        if cpp_file.endswith('.cu'):
            include_dirs.extend([
                '/usr/lib/openmpi/include',
                '/usr/lib/openmpi/include/openmpi'])

        if options is not None and 'include_dirs' in options.keys():
            include_dirs.extend(options['include_dirs'])

        command += ['-I' + incdir for incdir in include_dirs]
        command += ['-c', cpp_file, '-o', obj_file]
        print(' '.join(command))
        return subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def make_ext_from_objs(ext_file, obj_files, lib_dirs=None, libraries=None,
                       specials=None, options=None):

    cond = False
    if not os.path.exists(ext_file):
        cond = True
    else:
        date_ext = modification_date(ext_file)
        if any([date_ext < modification_date(obj_file)
                for obj_file in obj_files]):
            cond = True

    if cond:
        path_dir = os.path.split(ext_file)[0]
        if not os.path.exists(path_dir):
            os.makedirs(path_dir)

        # cxx = config_vars['CXX']
        ldshared = config_vars['LDSHARED']

        # keys = ['CONFIGURE_LDFLAGS']  # , 'CONFIGURE_CFLAGS']

        command = [w for w in ldshared.split()
                   if w not in ['-g']]

        if 'cufft' in ext_file:
            command = 'nvcc -Xcompiler -pthread -shared'.split()

        command += obj_files + ['-o', ext_file]
        if lib_dirs is not None:
            command.extend(['-L' + lib_dir for lib_dir in lib_dirs])

        if libraries is not None:
            command.extend(['-l' + lib for lib in libraries])

        print(' '.join(command))
        return subprocess.Popen(command)


def make_extensions(extensions, lib_dirs=None, libraries=None,
                    special=None, **options):

    if all(command not in sys.argv for command in [
            'build_ext', 'install', 'develop']):
        return

    if '--inplace' in sys.argv or 'develop' in sys.argv:
        path_base_output = '.'
    else:
        path_base_output = path_tmp

    sources = set()
    for ext in extensions:
        sources.update(ext.sources)

    # prepare a dictionary listing all files
    files = {}
    extension_files = ['pyx', 'cpp']

    for ext in extension_files:
        files[ext] = []

    for source in sources:
        for ext in extension_files:
            if source.endswith('.' + ext):
                files[ext].append(source)
        if source.endswith('.cu'):
            files['cpp'].append(source)
    files['o'] = []

    # cythonize .pyx files if needed
    for pyx_file in files['pyx']:
        cpp_file = os.path.splitext(pyx_file)[0] + '.cpp'

        full_module_name = None
        for ext in extensions:
            if pyx_file in ext.sources:
                full_module_name = ext.name

        result = make_cpp_from_pyx(
            cpp_file, pyx_file,
            full_module_name=full_module_name, options=options)

        files['cpp'].append(cpp_file)

    # compile .cpp files if needed
    processes = []
    for path in files['cpp']:
        result = os.path.join(path_tmp, os.path.splitext(path)[0] + '.o')
        p = make_obj_from_cpp(result, path, options)
        if p is not None:
            processes.append(p)
        files['o'].append(result)

    # wait for compilation
    while not all([process.poll() is not None for process in processes]):
        sleep(0.1)

    for p in processes:
        if p.returncode != 0:
            out, err = p.communicate()
            raise CompilationError(out, err)

    files['so'] = []

    # link .o files to produce the .so files if needed
    processes = []
    for ext in extensions:
        suffix = sysconfig.get_config_var('EXT_SUFFIX') or '.so'
        result = os.path.join(path_base_output,
                              ext.name.replace('.', os.path.sep) + suffix)
        objects = [
            os.path.join(path_tmp, os.path.splitext(source)[0] + '.o')
            for source in ext.sources]
        p = make_ext_from_objs(result, objects,
                               lib_dirs=lib_dirs, libraries=libraries)
        if p is not None:
            processes.append(p)
        files['so'].append(result)

    # wait for linking
    while not all([process.poll() is not None for process in processes]):
        sleep(0.1)

    # print(files)




class FFT2dFakeForDoc(object):
    """Class to perform Fast Fourier Transform in 2d."""
    def __init__(self, n0=2, n1=2):
        pass

    def get_local_size_X(self):
        pass

    def run_tests(self):
        """Run the c++ tests."""

    def run_benchs(self, nb_time_execute=10):
        """Run the c++ benchmarcks"""

    def fft_as_arg(self, fieldX, fieldK):
        """Perform the fft and copy the result in the second argument."""

    def ifft_as_arg(self, fieldK, fieldX):
        """Perform the ifft and copy the result in the second argument."""

    def fft(self, fieldX):
        """Perform the fft and returns the result."""

    def ifft(self, fieldK):
        """Perform the ifft and returns the result."""

    def get_shapeX_loc(self):
        pass

    def get_shapeK_loc(self):
        pass

    def get_shapeX_seq(self):
        pass

    def get_shapeK_seq(self):
        pass

    def get_is_transposed(self):
        """Get the boolean "is_transposed"."""

    def get_seq_indices_first_X(self):
        pass

    def get_seq_indices_first_K(self):
        pass

    def get_k_adim_loc(self):
        pass

    def get_x_adim_loc(self):
        pass

    def compute_energy_from_X(self, fieldX):
        pass

    def compute_energy_from_K(self, fieldK):
        pass




from __future__ import print_function, division

import unittest

import numpy as np

from math import pi

from fluiddyn.util import mpi

from fluidfft.fft2d import get_classes_seq, get_classes_mpi

n = 128

rank = mpi.rank
nb_proc = mpi.nb_proc

classes = get_classes_seq()

if nb_proc == 1:
    from fluidfft.fft2d.operator import OperatorPseudoSpectral2D
if nb_proc > 1:
    classes.update(get_classes_mpi())


def make_test_function(cls):

    def test(self):
        o = cls(n, n)
        o.run_tests()

        a = np.random.random(o.get_local_size_X()).reshape(
            o.get_shapeX_loc())
        afft = o.fft(a)
        a = o.ifft(afft)
        afft = o.fft(a)

        EX = o.compute_energy_from_X(a)
        EK = o.compute_energy_from_K(afft)

        self.assertTrue(EX == EK and EX != 0.)

        k0, k1 = o.get_k_adim_loc()

        if o.get_shapeK_loc() != (k0.size, k1.size):
            print(o.get_shapeK_loc(), k0.size, k1.size)
            raise Exception('o.get_shapeK_loc() != (k0.size, k1.size)')
    return test


def make_testop_function(name):

    def test(self):
        op = OperatorPseudoSpectral2D(25, 15, 3*pi, 1*pi)
        a = np.random.random(op._opfft.get_local_size_X()).reshape(
            op._opfft.get_shapeX_loc())
        afft = op.fft(a)
        a = op.ifft(afft)
        afft = op.fft(a)

        nrja = op.compute_energy_from_X(a)
        nrjafft = op.compute_energy_from_K(afft)
        self.assertEqual(nrja, nrjafft)

        nrjspa = (a**2).mean()/2

        energy_fft = 0.5 * abs(afft)**2

        E_kx, E_ky = op.compute_1dspectra(energy_fft)

        self.assertAlmostEqual(E_kx.sum()*op.deltakx, E_ky.sum()*op.deltaky)

        E_kh = op.compute_2dspectrum(energy_fft)

        self.assertAlmostEqual(nrjspa, E_kh.sum()*op.deltakh)
    return test


class Tests2D(unittest.TestCase):
    pass


for name, cls in classes.items():
    if cls is None:
        continue

    setattr(Tests2D, 'test_{0}'.format(name), make_test_function(cls))

if nb_proc == 1:
    setattr(Tests2D, 'test_operator2d', make_testop_function(name))


if __name__ == '__main__':
    unittest.main()



"""3d Fast Fourier Transform classes
====================================

This package contains different extension modules with classes for performing
Fast Fourier Transform with different methods and libraries. The number of
classes depend on how fluidfft has been compiled.

- `fluidfft3d.with_fftw3d.FFT3DWithFFTW3D`

- `fluidfft3d.with_cufft.FFT3DWithCUFFT`

- `fluidfft3d.mpi_with_fftwmpi3d.FFT3DMPIWithFFTWMPI3D`

- `fluidfft3d.mpi_with_pfft.FFT3DMPIWithPFFT`

- `fluidfft3d.mpi_with_p3dfft.FFT3DMPIWithP3DFFT` (do not work -> need .so)

All classes are very similar and provide the same public functions. Since these
classes are defined in Cython extensions that can not easily be compiled on the
readthedocs server, the API of the 3d classes is presented in this fake FFT
class:

.. autoclass:: FFT3dFakeForDoc
   :members:
   :undoc-members:

"""

from .. import import_fft_class

__all__ = ['FFT3dFakeForDoc', 'methods_seq', 'methods_mpi',
           'get_classes_seq', 'get_classes_mpi']

try:
    from .fake_mod_fft3d_for_doc import FFT3dFakeForDoc
except ImportError:
    pass

methods_seq = ['fftw3d']
methods_seq = ['fft3d.with_' + method for method in methods_seq]

methods_mpi = ['fftwmpi3d', 'p3dfft', 'pfft']
methods_mpi = ['fft3d.mpi_with_' + method for method in methods_mpi]


def get_classes_seq():
    return {method: import_fft_class(method, raise_import_error=False)
            for method in methods_seq}


def get_classes_mpi():
    return {method: import_fft_class(method, raise_import_error=False)
            for method in methods_mpi}




cdef extern from "base_fft.h":
    ctypedef struct mycomplex:
        pass

cdef extern from "${module_name}.h":
    cdef cppclass ${class_name}:
        int test()
        const char* bench(int)

        int get_local_size_X()
        int get_local_size_K()

        void get_local_shape_X(int*, int*)
        void get_local_shape_K(int*, int*)

        void get_shapeX_seq(int*, int*)
        void get_shapeK_seq(int*, int*)

        ${class_name}(int, int) except +

        void destroy()

        const char* get_classname()

        void fft(double* fieldX, mycomplex* fieldK)
        void ifft(mycomplex* fieldK, double* fieldX)

        double compute_energy_from_X(double* fieldX)
        double compute_energy_from_K(mycomplex* fieldK)
        double compute_mean_from_X(double* fieldX)
        double compute_mean_from_K(mycomplex* fieldK)

        char get_is_transposed()
        ptrdiff_t get_local_X0_start()
        ptrdiff_t get_local_K0_start()




include 'base.pyx'


from fft3dmpi_with_fftwmpi3d_pxd cimport (
    FFT3DMPIWithFFTWMPI3D as mycppclass,
    mycomplex)


cdef class FFT3DMPIWithFFTWMPI3D:
    """Perform fast Fourier transform in 3D.

    """
    cdef mycppclass* thisptr
    cdef tuple _shapeK_loc, _shapeX_loc
    cdef public MPI.Comm comm
    cdef public int nb_proc, rank

    def __cinit__(self, int n0=2, int n1=2, int n2=4):
        self.thisptr = new mycppclass(n0, n1, n2)

    def __init__(self, int n0=2, int n1=2, int n2=4):
        self._shapeK_loc = self.get_shapeK_loc()
        self._shapeX_loc = self.get_shapeX_loc()
        # info on MPI
        self.nb_proc = nb_proc
        self.rank = rank
        if self.nb_proc > 1:
            self.comm = comm

    def __dealloc__(self):
        self.thisptr.destroy()
        del self.thisptr

    def get_local_size_X(self):
        return self.thisptr.get_local_size_X()

    def run_tests(self):
        return self.thisptr.test()

    def run_benchs(self, nb_time_execute=10):
        txt = self.thisptr.bench(nb_time_execute)
        return tuple(float(word) for word in txt.split() if word[0].isdigit())

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef fft_generic(self, DTYPEf_t[:, :, ::1] fieldX,
                      DTYPEc_t[:, :, ::1] fieldK=None):
        if fieldK is None:
            fieldK = np.empty(self._shapeK_loc, dtype=DTYPEc, order='C')
        self.thisptr.fft(&fieldX[0, 0, 0], <mycomplex*> &fieldK[0, 0, 0])
        return np.array(fieldK)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef ifft_generic(self, DTYPEc_t[:, :, ::1] fieldK,
                       DTYPEf_t[:, :, ::1] fieldX=None):
        if fieldX is None:
            fieldX = np.empty(self._shapeX_loc, dtype=DTYPEf, order='C')
        self.thisptr.ifft(<mycomplex*> &fieldK[0, 0, 0], &fieldX[0, 0, 0])
        return np.array(fieldX)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef fft_as_arg(self, DTYPEf_t[:, :, ::1] fieldX,
                     DTYPEc_t[:, :, ::1] fieldK):
        self.thisptr.fft(&fieldX[0, 0, 0], <mycomplex*> &fieldK[0, 0, 0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef ifft_as_arg(self, DTYPEc_t[:, :, ::1] fieldK,
                      DTYPEf_t[:, :, ::1] fieldX):
        self.thisptr.ifft(<mycomplex*> &fieldK[0, 0, 0], &fieldX[0, 0, 0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef fft(self, DTYPEf_t[:, :, ::1] fieldX):
        cdef np.ndarray[DTYPEc_t, ndim=3] fieldK
        fieldK = np.empty(self.get_shapeK_loc(), dtype=DTYPEc, order='C')
        self.thisptr.fft(&fieldX[0, 0, 0], <mycomplex*> &fieldK[0, 0, 0])
        return fieldK

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef ifft(self, DTYPEc_t[:, :, ::1] fieldK):
        cdef np.ndarray[DTYPEf_t, ndim=3] fieldX
        fieldX = np.empty(self.get_shapeX_loc(), dtype=DTYPEf, order='C')
        self.thisptr.ifft(<mycomplex*> &fieldK[0, 0, 0], &fieldX[0, 0, 0])
        return fieldX

    cpdef get_shapeX_loc(self):
        """Get the shape of the array in real space for this mpi process."""
        cdef int nX0loc, nX1loc, nX2
        self.thisptr.get_local_shape_X(&nX0loc, &nX1loc, &nX2)
        return nX0loc, nX1loc, nX2

    cpdef get_shapeK_loc(self):
        """Get the shape of the array in Fourier space for this mpi process."""
        cdef int nK0loc, nK1loc, nK2
        self.thisptr.get_local_shape_K(&nK0loc, &nK1loc, &nK2)
        return nK0loc, nK1loc, nK2

    cpdef get_shapeX_seq(self):
        """Get the shape of an array in real space for a sequential run."""
        cdef int nX0, nX1, nX2
        self.thisptr.get_global_shape_X(&nX0, &nX1, &nX2)
        return nX0, nX1, nX2

    def gather_Xspace(self, DTYPEf_t[:, :, ::1] ff_loc,
                      root=None):
        """Gather an array in real space for a parallel run."""
        cdef np.ndarray[DTYPEf_t, ndim=3] ff_seq

        # self.shapeX_loc is the same for all processes,
        # it is safe to use Allgather or Gather
        if root is None:
            ff_seq = np.empty(self.get_shapeX_seq(), DTYPEf)
            self.comm.Allgather(ff_loc, ff_seq)
        elif isinstance(root, int):
            ff_seq = None
            if self.rank == root:
                ff_seq = np.empty(self.get_shapeX_seq(), DTYPEf)
            self.comm.Gather(ff_loc, ff_seq, root=root)
        else:
            raise ValueError('root should be an int')
        return ff_seq

    def scatter_Xspace(self, DTYPEf_t[:, :, ::1] ff_seq,
                      root=None):
        """Gather an array in real space for a parallel run."""
        cdef np.ndarray[DTYPEf_t, ndim=3] ff_loc

        # self.shapeX_loc is the same for all processes,
        # it is safe to use Allgather or Gather
        if root is None:
            ff_loc = np.empty(self.get_shapeX_loc(), DTYPEf)
            self.comm.Scatter(ff_seq, ff_loc, root=0)
        elif isinstance(root, int):
            ff_loc = None
            if self.rank == root:
                ff_loc = np.empty(self.get_shapeX_loc(), DTYPEf)
            self.comm.Scatter(ff_seq, ff_loc, root=root)
        else:
            raise ValueError('root should be an int')
        return ff_loc

    cpdef get_shapeK_seq(self):
        """Get the shape of an array in Fourier space for a sequential run."""
        cdef int nK0, nK1, nK2
        self.thisptr.get_global_shape_K(&nK0, &nK1, &nK2)
        return nK0, nK1, nK2

    def sum_wavenumbers(self, fieldK):
        """Compute the sum over all wavenumbers."""
        if fieldK.dtype == np.float64:
            return self._sum_wavenumbers_double(fieldK)
        elif fieldK.dtype == np.complex128:
            return self._sum_wavenumbers_complex(fieldK)
        else:
            raise TypeError('dtype of fieldK has to be float64 or complex128.')

    cdef _sum_wavenumbers_double(self, DTYPEf_t[:,:,::1] fieldK):
        return self.thisptr.sum_wavenumbers_double(
            <DTYPEf_t*> &fieldK[0, 0, 0])

    cdef _sum_wavenumbers_complex(self, DTYPEc_t[:,:,::1] fieldK):
        cdef DTYPEc_t result
        self.thisptr.sum_wavenumbers_complex(
            <mycomplex*> &fieldK[0, 0, 0], <mycomplex*> &result)
        return result

    cpdef get_dimX_K(self):
        """Get the indices of the real space dimension in Fourier space."""
        cdef int d0, d1, d2
        self.thisptr.get_dimX_K(&d0, &d1, &d2)
        return d0, d1, d2

    cpdef get_seq_indices_first_K(self):
        """Get the "sequential" index of the first number in Fourier space."""
        cdef int i0, i1
        self.thisptr.get_seq_indices_first_K(&i0, &i1)
        return i0, i1

    cpdef get_k_adim_loc(self):
        """Get the non-dimensional wavenumbers stored locally.

        returns k0_adim_loc, k1_adim_loc, k2_adim_loc.

        """
        cdef int nK0, nK1, nK2, nK0_loc, nK1_loc, nK2_loc
        cdef int d0, d1, d2, i0_start, i1_start
        cdef np.ndarray tmp, k0_adim_loc, k1_adim_loc, k2_adim_loc

        nK0, nK1, nK2 = self.get_shapeK_seq()
        nK0_loc, nK1_loc, nK2_loc = self.get_shapeK_loc()

        d0, d1, d2 = self.get_dimX_K()
        i0_start, i1_start = self.get_seq_indices_first_K()

        if d0 == 2:
            tmp = np.arange(nK0)
        else:
            k0_adim_max = nK0//2
            tmp = np.r_[0:k0_adim_max+1, -k0_adim_max+1:0]

        k0_adim_loc = tmp[i0_start:i0_start+nK0_loc]

        if d1 == 2:
            tmp = np.arange(nK1)
        else:
            k1_adim_max = nK1//2
            tmp = np.r_[0:k1_adim_max+1, -k1_adim_max+1:0]

        k1_adim_loc = tmp[i1_start:i1_start+nK1_loc]

        if d2 == 2:
            k2_adim_loc = np.arange(nK2)
        else:
            k2_adim_max = nK2//2
            k2_adim_loc = np.r_[0:k2_adim_max+1, -k2_adim_max+1:0]

        return k0_adim_loc, k1_adim_loc, k2_adim_loc

    def build_invariant_arrayX_from_2d_indices12X(self, o2d, arr2d):

        nX0, nX1, nX2 = self.get_shapeX_seq()
        nX0loc, nX1loc, nX2loc = self.get_shapeX_loc()

        if (nX1, nX2) != tuple(o2d.get_shapeX_seq()):
            raise ValueError('Not the same physical shape...')

        # check that the 2d fft is not with distributed memory...
        if tuple(o2d.get_shapeX_loc()) != tuple(o2d.get_shapeX_loc()):
            raise ValueError('2d fft is with distributed memory...')

        ind0seq_first, ind1seq_first = self.get_seq_indices_first_K()

        if (nX1loc, nX2loc) == o2d.get_shapeX_loc():
            arr3d_loc_2dslice = arr2d
        else:
            raise NotImplementedError

        arr3d = np.empty([nX0loc, nX1loc, nX2loc])
        for i0 in range(nX0loc):
            arr3d[i0] = arr3d_loc_2dslice

        return arr3d

    def build_invariant_arrayK_from_2d_indices12X(self, o2d, arr2d):

        nK0, nK1, nK2 = self.get_shapeK_seq()
        nK0loc, nK1loc, nK2loc = self.get_shapeK_loc()

        nX0, nX1, nX2 = self.get_shapeX_seq()

        if (nX1, nX2) != o2d.get_shapeX_seq():
            raise ValueError('Not the same physical shape...')

        # check that the 2d fft is not with distributed memory...
        if o2d.get_shapeX_loc() != o2d.get_shapeX_loc():
            raise ValueError('2d fft is with distributed memory...')

        ind0seq_first, ind1seq_first = self.get_seq_indices_first_K()
        dimX_K = self.get_dimX_K()

        arr3d = np.zeros([nK0loc, nK1loc, nK2loc], dtype=np.complex128)

        if dimX_K == (0, 1, 2):
            # simple
            if (nK0, nK1, nK2) == (nK0loc, nK1loc, nK2loc):
                # very simple
                arr3d_loc_2dslice = arr2d
            else:
                raise NotImplementedError

            arr3d[0] = arr3d_loc_2dslice

        elif dimX_K == (1, 0, 2):
            # like fft3d.mpi_with_fftwmpi3d
            arr3d_loc_2dslice = np.zeros([nK0loc, nK2loc], dtype=np.complex128)

            for i0 in range(nK0loc):
                for i2 in range(nK2loc):
                    i0_2d = ind0seq_first + i0
                    i1_2d = i2
                    arr3d_loc_2dslice[i0, i2] = arr2d[i0_2d, i1_2d]

            arr3d[:, 0, :] = arr3d_loc_2dslice
        else:
            raise NotImplementedError

        return arr3d


FFTclass = FFT3DMPIWithFFTWMPI3D




import numpy as np

from fluidfft import import_fft_class

FFT3D = import_fft_class('fft3d.mpi_with_fftwmpi3d')
FFT2D = import_fft_class('fft2d.with_fftw2d')


def build_arrayX_from_2d_indices12(self, o2d, arr2d):

    nX0, nX1, nX2 = self.get_shapeX_seq()
    nX0loc, nX1loc, nX2loc = self.get_shapeX_loc()

    if (nX1, nX2) != o2d.get_shapeX_seq():
        raise ValueError('Not the same physical shape...')

    # check that the 2d fft is not with distributed memory...
    if o2d.get_shapeX_loc() != o2d.get_shapeX_loc():
        raise ValueError('2d fft is with distributed memory...')

    if (nX1loc, nX2loc) == o2d.get_shapeX_loc():
        arr3d_loc_2dslice = arr2d
    else:
        raise NotImplementedError

    arr3d = np.empty([nX0loc, nX1loc, nX2loc])
    for i0 in range(nX0loc):
        arr3d[i0] = arr3d_loc_2dslice

    return arr3d


def build_invariant_arrayK_from_2d_indices12X(self, o2d, arr2d):

    nK0, nK1, nK2 = self.get_shapeK_seq()
    nK0loc, nK1loc, nK2loc = self.get_shapeK_loc()

    nX0, nX1, nX2 = self.get_shapeX_seq()

    if (nX1, nX2) != o2d.get_shapeX_seq():
        raise ValueError('Not the same physical shape...')

    # check that the 2d fft is not with distributed memory...
    if o2d.get_shapeX_loc() != o2d.get_shapeX_loc():
        raise ValueError('2d fft is with distributed memory...')

    ind0seq_first, ind1seq_first = self.get_seq_indices_first_K()
    dimX_K = self.get_dimX_K()

    arr3d = np.zeros([nK0loc, nK1loc, nK2loc], dtype=np.complex128)

    if dimX_K == (0, 1, 2):
        # simple
        if (nK0, nK1, nK2) == (nK0loc, nK1loc, nK2loc):
            # very simple
            arr3d_loc_2dslice = arr2d
        else:
            raise NotImplementedError

        arr3d[0] = arr3d_loc_2dslice

    elif dimX_K == (1, 0, 2):
        # like fft3d.mpi_with_fftwmpi3d
        arr3d_loc_2dslice = np.zeros([nK0loc, nK2loc], dtype=np.complex128)

        for i0 in range(nK0loc):
            for i2 in range(nK2loc):
                i0_2d = ind0seq_first + i0
                i1_2d = i2
                arr3d_loc_2dslice[i0, i2] = arr2d[i0_2d, i1_2d]

        arr3d[:, 0, :] = arr3d_loc_2dslice
    else:
        raise NotImplementedError

    return arr3d


n = 4

n0 = n
n1 = 10*n
n2 = 100*n

o3d = FFT3D(n0, n1, n2)
o2d = FFT2D(n1, n2)

shapeK_loc_2d = o2d.get_shapeK_loc()

arr2d = np.arange(np.product(shapeK_loc_2d)).reshape(
    shapeK_loc_2d).astype(np.complex128)

arr3d = build_invariant_arrayK_from_2d_indices12X(o3d, o2d, arr2d)
# arr3d = o3d.build_invariant_arrayK_from_2d_indices12X(o2d, arr2d)




from __future__ import print_function

import os
from runpy import run_path

from setuptools import setup, find_packages

from purepymake import Extension, make_extensions
from config import get_config
from src_cy.make_files_with_mako import make_pyx_files

make_pyx_files()

config = get_config()

# Get the long description from the relevant file
with open('README.rst') as f:
    long_description = f.read()
lines = long_description.splitlines(True)
long_description = ''.join(lines[12:])

# Get the version from the relevant file
d = run_path('fluidfft/_version.py')
__version__ = d['__version__']

# make a python module from cython files
run_path('src_cy/create_fake_mod_for_doc.py')

src_cpp_dir = 'src_cpp'
src_cy_dir = 'src_cy'
src_base = 'src_cpp/base'
src_cpp_3d = 'src_cpp/3d'
src_cpp_2d = 'src_cpp/2d'


def create_ext(base_name):

    if base_name.startswith('fft2d'):
        dim = '2d'
    elif base_name.startswith('fft3d'):
        dim = '3d'
    else:
        raise ValueError()

    src_cpp_dim = os.path.join(src_cpp_dir, dim)

    source_ends = ['_cy.pyx']
    if base_name.endswith('cufft'):
        source_ends.append('.cu')
    else:
        source_ends.append('.cpp')

    source_files = [base_name + end for end in source_ends]

    base_name = base_name[len('fft2d'):]
    if base_name.startswith('_'):
        base_name = base_name[1:]

    sources = []
    for name_file in source_files:
        if name_file.endswith('_cy.pyx'):
            path = os.path.join(src_cy_dir, name_file)
        else:
            path = os.path.join(src_cpp_dim, name_file)
        sources.append(path)

    sources.extend(
        [os.path.join(src_base, 'base_fft.cpp'),
         os.path.join(src_cpp_dim, 'base_fft' + dim + '.cpp')])
    if base_name.startswith('mpi'):
        sources.extend(
            [os.path.join(src_base, 'base_fftmpi.cpp'),
             os.path.join(src_cpp_dim, 'base_fft' + dim + 'mpi.cpp')])

    return Extension(
        name='fluidfft.fft' + dim + '.' + base_name,
        sources=sources)


base_names = []
if config['fftw']['use']:
        base_names.extend([
            'fft2d_with_fftw1d', 'fft2d_with_fftw2d', 'fft2dmpi_with_fftw1d',
            'fft3d_with_fftw3d'])

if config['fftw-mpi']['use']:
        base_names.extend([
            'fft2dmpi_with_fftwmpi2d', 'fft3dmpi_with_fftwmpi3d'])

if config['cufft']['use']:
    base_names.extend(['fft2d_with_cufft'])
    base_names.extend(['fft3d_with_cufft'])

if config['pfft']['use']:
    base_names.extend(['fft3dmpi_with_pfft'])

if config['p3dfft']['use']:
    base_names.extend(['fft3dmpi_with_p3dfft'])


on_rtd = os.environ.get('READTHEDOCS')
if on_rtd:
    base_names = []
else:
    import mpi4py
    if mpi4py.__version__[0] < '2':
        raise ValueError('Please upgrade to mpi4py >= 2.0')

    import numpy as np

    ext_modules = []
    libraries = set(['fftw3', 'mpi_cxx'])
    lib_dirs = set()
    include_dirs = set(
        [src_cy_dir, src_base, src_cpp_3d, src_cpp_2d,
         'include', mpi4py.get_include(), np.get_include()])

    specials = {}


for base_name in base_names:
    ext_modules.append(create_ext(base_name))
    TMP = os.getenv('FFTW3_INC_DIR')
    if TMP is not None:
        include_dirs.add(TMP)
    TMP = os.getenv('FFTW3_LIB_DIR')
    if TMP is not None:
        lib_dirs.add(TMP)
    if 'fftwmpi' in base_name:
        libraries.add('fftw3_mpi')
    elif 'pfft' in base_name:
        libraries.update(['fftw3_mpi', 'pfft'])
#        include_dirs.update(['/opt/pfft/1.0.6/include'])
        lib_dirs.update(['/opt/pfft/1.0.6/lib'])
    elif 'p3dfft' in base_name:
        libraries.update(['p3dfft'])
#        include_dirs.update(['/opt/p3dfft/2.7.4-mt/include'])
#        lib_dirs.update(['/opt/p3dfft/2.7.4-mt/lib'])
    elif 'cufft' in base_name:
        libraries.add('cufft')
        include_dirs.update(['/opt/cuda/7.5/include'])
        lib_dirs.update(['/opt/cuda/7.5/lib64'])
        specials[''] = {'CC': 'nvcc'}


if not on_rtd:
    make_extensions(
        ext_modules, include_dirs=include_dirs,
        lib_dirs=lib_dirs, libraries=libraries,
        specials=specials, CC='mpicxx')


setup(
    name='fluidfft',
    version=__version__,
    description=('Efficient and easy Fast Fourier Transform for Python.'),
    long_description=long_description,
    keywords='Fast Fourier Transform, spectral code',
    author='Pierre Augier',
    author_email='pierre.augier@legi.cnrs.fr',
    url='https://bitbucket.org/fluiddyn/fluidfft',
    license='CeCILL',
    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        # actually CeCILL License (GPL compatible license for French laws)
        #
        # Specify the Python versions you support here. In particular,
        # ensure that you indicate whether you support Python 2,
        # Python 3 or both.
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Cython',
        'Programming Language :: C'],
    packages=find_packages(exclude=[
        'doc', 'include', 'scripts', 'src_cpp', 'src_cy']),
    install_requires=['fluiddyn', 'mako']
    # cmdclass={'build_ext': build_ext},
    # ext_modules=ext_modules
)




from __future__ import print_function

import numpy as np

from math import pi, sqrt

from builtins import range

from fluiddyn.util import mpi

from fluidfft import create_fft_object

if mpi.nb_proc > 1:
    raise NotImplementedError


class OperatorPseudoSpectral2D(object):

    def __init__(self, nx, ny, lx, ly, fft='fft2d.with_fftw2d',
                 coef_dealiasing=1.):

        self.nx = nx = int(nx)
        self.ny = ny = int(ny)
        self.lx = lx = float(lx)
        self.ly = ly = float(ly)

        if isinstance(fft, str):
            print(fft)
            opfft = create_fft_object(fft, ny, nx)
        else:
            opfft = fft

        self._opfft = opfft

        self.is_transposed = opfft.get_is_transposed()

        if self.is_transposed:
            raise NotImplementedError

        self.fft = self._opfft.fft
        self.ifft = self._opfft.ifft
        self.fft_as_arg = self._opfft.fft_as_arg
        self.ifft_as_arg = self._opfft.ifft_as_arg
        self.shapeX_loc = self._opfft.get_shapeX_loc()
        self.shapeX_seq = self._opfft.get_shapeX_seq()
        self.shapeK_loc = self._opfft.get_shapeK_loc()
        self.shapeK_seq = self._opfft.get_shapeK_seq()
        self.compute_energy_from_X = self._opfft.compute_energy_from_X
        self.compute_energy_from_K = self._opfft.compute_energy_from_K

        self.dx = lx/nx
        self.dy = ly/ny
        self.x = self.dx * np.arange(nx)
        self.y = self.dy * np.arange(ny)

        self.deltakx = 2*pi/lx
        self.deltaky = 2*pi/ly

        k0_adim, k1_adim = opfft.get_k_adim_loc()

        # true only is not transposed...
        self.kx = self.deltakx * k1_adim
        self.ky = self.deltaky * k0_adim

        self.kx_loc = self.kx
        self.ky_loc = self.ky

        if not self.is_transposed:
            [self.KX, self.KY] = np.meshgrid(self.kx_loc, self.ky_loc)
            self.dim_kx = 1
            self.dim_ky = 0
            self.k0 = self.ky
            self.k1 = self.kx
        else:
            [self.KY, self.KX] = np.meshgrid(self.ky_loc, self.kx_loc)
            self.dim_kx = 0
            self.dim_ky = 1
            self.k0 = self.kx
            self.k1 = self.ky

        assert self.KX.shape == self.shapeK_loc

        self.KX2 = self.KX**2
        self.KY2 = self.KY**2
        self.K2 = self.KX2 + self.KY2
        self.K4 = self.K2**2
        self.K8 = self.K4**2
        self.KK = np.sqrt(self.K2)

        self.nkx_seq = nx//2 + 1
        self.nky_seq = ny
        self.nky_spectra = ny//2 + 1

        khmax = min(self.kx.max(), self.ky.max())
        self.deltakh = max(self.deltakx, self.deltaky)
        self.nkh = int(khmax / self.deltakh)
        if self.nkh == 0:
            self.nkh = 1
        self.kh_2dspectrum = self.deltakh*np.arange(self.nkh)

        # Initialisation dealiasing
        self.coef_dealiasing = coef_dealiasing
        kx_max = self.deltakx * (nx//2 + 1)
        ky_max = self.deltaky * (ny//2 + 1)
        CONDKX = abs(self.KX) > coef_dealiasing*kx_max
        CONDKY = abs(self.KY) > coef_dealiasing*ky_max
        where_dealiased = np.logical_or(CONDKX, CONDKY)
        self.where_dealiased = np.array(where_dealiased, dtype=np.uint8)

    def compute_1dspectra(self, energy_fft):
        if mpi.nb_proc == 1 and not self.is_transposed:
            # In this case, self.dim_ky == 0 and self.dim_ky == 1
            # Memory is not shared
            # note that only the kx >= 0 are in the spectral variables

            # the 2 is here because there are only the kx >= 0
            energy_fft = energy_fft.copy()

            n_tmp = self.nkx_seq
            if self.nx % 2 == 0:
                n_tmp -= 1

            energy_fft[:, 1:n_tmp] *= 2

            # to obtain the spectrum as a function of kx
            # we sum over all ky
            E_kx = energy_fft.sum(0)/self.deltakx

            # computation of E_ky
            E_ky_temp = energy_fft.sum(1)
            nkyE = self.nky_spectra
            E_ky = E_ky_temp[0:nkyE]
            n_tmp = nkyE
            if self.ny % 2 == 0:
                n_tmp -= 1
            E_ky[1:n_tmp] += E_ky_temp[self.nky_seq:nkyE-1:-1]
            E_ky /= self.deltaky

            return E_kx, E_ky
        else:
            raise NotImplementedError

    def compute_2dspectrum(self, energy_fft):
        if mpi.nb_proc == 1 and not self.is_transposed:
            # In this case, self.dim_ky == 0 and self.dim_ky == 1
            # Memory is not shared
            # note that only the kx >= 0 are in the spectral variables

            # the 2 is here because there are only the kx >= 0
            energy_fft = energy_fft.copy()

            n_tmp = self.nkx_seq
            if self.nx % 2 == 0:
                n_tmp -= 1

            energy_fft[:, 1:n_tmp] *= 2

            E_kh = np.zeros_like(self.kh_2dspectrum)
            n0, n1 = self.shapeK_loc
            ikhmax = len(self.kh_2dspectrum) - 1

            for i0 in range(n0):
                k0 = self.k0[i0]
                for i1 in range(n1):
                    k1 = self.k1[i1]
                    ikh = int(sqrt(k0**2 + k1**2)/self.deltakh)
                    if ikh > ikhmax:
                        ikh = ikhmax
                    E_kh[ikh] += energy_fft[i0, i1]

            return E_kh/self.deltakh
        else:
            raise NotImplementedError


if __name__ == '__main__':
    self = OperatorPseudoSpectral2D(5, 3, 2*pi, 1*pi)

    a = np.random.random(self._opfft.get_local_size_X()).reshape(
        self._opfft.get_shapeX_loc())
    afft = self.fft(a)
    a = self.ifft(afft)
    afft = self.fft(a)

    print('energy spatial C:', self.compute_energy_from_X(a))
    print('energy fft      :', self.compute_energy_from_K(afft))

    print('energy spatial P:', (a**2).mean()/2)

    energy_fft = 0.5 * abs(afft)**2

    E_kx, E_ky = self.compute_1dspectra(energy_fft)

    print('energy E_kx     ;', E_kx.sum()*self.deltakx)
    print('energy E_ky     :', E_ky.sum()*self.deltaky)

    E_kh = self.compute_2dspectrum(energy_fft)

    print('energy E_kh     :', E_kh.sum()*self.deltakh)




cdef extern from "base_fft.h":
    ctypedef struct mycomplex:
        pass

cdef extern from "fft3dmpi_with_pfft.h":
    cdef cppclass FFT3DMPIWithPFFT:
        int test()
        const char* bench(int)

        int get_local_size_X()
        int get_local_size_K()

        void get_local_shape_X(int*, int*, int*)
        void get_local_shape_K(int*, int*, int*)

        void get_global_shape_X(int*, int*, int*)
        void get_global_shape_K(int*, int*, int*)

        void get_dimX_K(int*, int*, int*)
        void get_seq_indices_first_K(int*, int*)
        
        FFT3DMPIWithPFFT(int, int, int) except +

        void destroy()

        const char* get_classname()

        void fft(double* fieldX, mycomplex* fieldK)
        void ifft(mycomplex* fieldK, double* fieldX)

        double sum_wavenumbers_double(double* fieldK)
        void sum_wavenumbers_complex(mycomplex* fieldK, mycomplex* result)
        
        double compute_energy_from_X(double* fieldX)
        double compute_energy_from_K(mycomplex* fieldK)
        double compute_mean_from_X(double* fieldX)
        double compute_mean_from_K(mycomplex* fieldK)



"""Do not modify this file to modify your configuration. Copy
site.cfg.example to site.cfg and modify this file.

"""

import os

try:  # python 3
    from configparser import ConfigParser
except:  # python 2.7
    from ConfigParser import ConfigParser


def get_default_config():

    config = ConfigParser()

    sections = ['fftw', 'fftw-mpi', 'cufft', 'pfft', 'p3dfft']

    use = {k: 'False' for k in sections}
    use['fftw'] = 'True'

    for section in sections:
        config.add_section(section)
        config.set(section, 'use', use[section])
        config.set(section, 'dir', '')
        config.set(section, 'include_dir', '')
        config.set(section, 'library_dir', '')

    return config


def make_site_cfg_example_file():

    config = get_default_config()

    with open('site.cfg.example', 'wb') as configfile:
        config.write(configfile)


def get_config():
    config = get_default_config()

    if os.path.exists('site.cfg'):
        config.read('site.cfg')

    config_dict = {}
    for section in config.sections():
        section_dict = {}
        for option in config.options(section):
            value = config.get(section, option)
            if option == 'use':
                value = value.lower()
                if value not in ['true', 'false']:
                    raise ValueError('"use" should be "True" of "False".')
                value = value == 'true'
            section_dict[option] = value
        config_dict[section] = section_dict

    return config_dict


if __name__ == '__main__':
    make_site_cfg_example_file()




include 'base.pyx'


from ${module_name}_pxd cimport (
    ${class_name} as mycppclass,
    mycomplex)


cdef class ${class_name}:
    """Class to perform Fast Fourier Transform in 2d."""
    cdef mycppclass* thisptr
    cdef tuple _shape_K_loc, _shape_X_loc

    def __cinit__(self, int n0=2, int n1=2):
        self.thisptr = new mycppclass(n0, n1)

    def __init__(self, int n0=2, int n1=2):
        self._shape_K_loc = self.get_shapeK_loc()
        self._shape_X_loc = self.get_shapeX_loc()

    def __dealloc__(self):
        self.thisptr.destroy()
        del self.thisptr

    def get_local_size_X(self):
        return self.thisptr.get_local_size_X()
    
    def run_tests(self):
        """Run the c++ tests."""
        return self.thisptr.test()

    def run_benchs(self, nb_time_execute=10):
        """Run the c++ benchmarcks"""
        txt = self.thisptr.bench(nb_time_execute)
        return tuple(float(word) for word in txt.split() if word[0].isdigit())

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef fft_as_arg(self, DTYPEf_t[:, ::1] fieldX,
                     DTYPEc_t[:, ::1] fieldK):
        """Perform the fft and copy the result in the second argument."""
        self.thisptr.fft(&fieldX[0, 0], <mycomplex*> &fieldK[0, 0])
        return fieldK
        
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef ifft_as_arg(self, DTYPEc_t[:, ::1] fieldK,
                      DTYPEf_t[:, ::1] fieldX):
        """Perform the ifft and copy the result in the second argument."""
        self.thisptr.ifft(<mycomplex*> &fieldK[0, 0], &fieldX[0, 0])
        return fieldX

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef fft(self, DTYPEf_t[:, ::1] fieldX):
        """Perform the fft and returns the result."""
        cdef np.ndarray[DTYPEc_t, ndim=2] fieldK
        fieldK = np.empty(self.get_shapeK_loc(), dtype=DTYPEc, order='C')
        self.thisptr.fft(&fieldX[0, 0], <mycomplex*> &fieldK[0, 0])
        return fieldK

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef ifft(self, DTYPEc_t[:, ::1] fieldK):
        """Perform the ifft and returns the result."""
        cdef np.ndarray[DTYPEf_t, ndim=2] fieldX
        fieldX = np.empty(self.get_shapeX_loc(), dtype=DTYPEf, order='C')
        self.thisptr.ifft(<mycomplex*> &fieldK[0, 0], &fieldX[0, 0])
        return fieldX

    def get_shapeX_loc(self):
        cdef int output_nX0loc, output_nX1
        self.thisptr.get_local_shape_X(&output_nX0loc, &output_nX1)
        return output_nX0loc, output_nX1

    def get_shapeK_loc(self):
        cdef int output_nK0loc, output_nK1
        self.thisptr.get_local_shape_K(&output_nK0loc, &output_nK1)
        return output_nK0loc, output_nK1

    def get_shapeX_seq(self):
        cdef int output_nX0, output_nX1
        self.thisptr.get_shapeX_seq(&output_nX0, &output_nX1)
        return output_nX0, output_nX1

    def get_shapeK_seq(self):
        cdef int output_nK0, output_nK1
        self.thisptr.get_shapeK_seq(&output_nK0, &output_nK1)
        return output_nK0, output_nK1

    def get_is_transposed(self):
        """Get the boolean "is_transposed"."""
        return bool(self.thisptr.get_is_transposed())

    def get_seq_indices_first_X(self):
        return <int> self.thisptr.get_local_X0_start(), 0

    def get_seq_indices_first_K(self):
        return <int> self.thisptr.get_local_K0_start(), 0

    def get_k_adim_loc(self):
        nyseq, nxseq = self.get_shapeX_seq()

        kyseq = np.array(list(range(nyseq//2 + 1)) + list(range(-nyseq//2 + 1, 0)))
        kxseq = np.array(range(nxseq//2 + 1))

        if self.get_is_transposed():
            k0seq, k1seq = kxseq, kyseq
        else:
            k0seq, k1seq = kyseq, kxseq

        ik0_start, ik1_start = self.get_seq_indices_first_K()
        nk0loc, nk1loc = self.get_shapeK_loc()

        k0loc = k0seq[ik0_start:ik0_start+nk0loc]
        k1loc = k1seq[ik1_start:ik1_start+nk1loc]

        return k0loc, k1loc

    def get_x_adim_loc(self):
        nyseq, nxseq = self.get_shapeX_seq()

        ix0_start, ix1_start = self.get_seq_indices_first_X()
        nx0loc, nx1loc = self.get_shapeX_loc()

        x0loc = np.array(range(ix0_start, ix0_start+nx0loc))
        x1loc = np.array(range(ix1_start, ix1_start+nx1loc))

        return x0loc, x1loc

    def compute_energy_from_X(self, DTYPEf_t[:, ::1] fieldX):
        return <float> self.thisptr.compute_energy_from_X(&fieldX[0, 0])

    def compute_energy_from_K(self, DTYPEc_t[:, ::1] fieldK):
        return <float> self.thisptr.compute_energy_from_K(
            <mycomplex*> &fieldK[0, 0])

    

FFTclass = ${class_name}




class FFT3dFakeForDoc(object):
    """Perform fast Fourier transform in 3D.

    """
    def __init__(self, n0=2, n1=2, n2=4):
        pass

    def get_local_size_X(self):
        pass

    def run_tests(self):
        pass

    def run_benchs(self, nb_time_execute=10):
        pass

    def fft_generic(self, fieldX, fieldK=None):
        pass

    def ifft_generic(self, fieldK, fieldX=None):
        pass

    def fft_as_arg(self, fieldX, fieldK):
        pass

    def ifft_as_arg(self, fieldK, fieldX):
        pass

    def fft(self, fieldX):
        pass

    def ifft(self, fieldK):
        pass

    def get_shapeX_loc(self):
        """Get the shape of the array in real space for this mpi process."""

    def get_shapeK_loc(self):
        """Get the shape of the array in Fourier space for this mpi process."""

    def get_shapeX_seq(self):
        """Get the shape of an array in real space for a sequential run."""

    def gather_Xspace(self, ff_loc, root=None):
        """Gather an array in real space for a parallel run."""

    def scatter_Xspace(self, ff_seq, root=None):
        """Gather an array in real space for a parallel run."""

    def get_shapeK_seq(self):
        """Get the shape of an array in Fourier space for a sequential run."""

    def sum_wavenumbers(self, fieldK):
        """Compute the sum over all wavenumbers."""

    def get_dimX_K(self):
        """Get the indices of the real space dimension in Fourier space."""

    def get_seq_indices_first_K(self):
        """Get the "sequential" index of the first number in Fourier space."""

    def get_k_adim_loc(self):
        """Get the non-dimensional wavenumbers stored locally.

        returns k0_adim_loc, k1_adim_loc, k2_adim_loc.

        """

    def build_invariant_arrayX_from_2d_indices12X(self, o2d, arr2d):
        pass

    def build_invariant_arrayK_from_2d_indices12X(self, o2d, arr2d):
        pass



"""Simple Fast Fourier Transform for Python
===========================================

This module provides two helper functions to import fft classes and
create fft objects:

.. autofunction:: import_fft_class

.. autofunction:: create_fft_object

The fft classes are in the two subpackages

.. autosummary::
   :toctree:

   fft2d
   fft3d

"""

from __future__ import print_function

from fluidfft._version import __version__

from importlib import import_module as _import_module

__all__ = ['__version__', 'import_fft_class', 'create_fft_object']


def import_fft_class(method, raise_import_error=True):
    """Import a fft class.

    Parameters
    ----------

    method : str
      Name of module or string characterizing a method. It has to correspond to
      a module of fluidfft. The first part "fluidfft." of the module "path" can
      be omitted.

    raise_import_error : {True}, False

      If raise_import_error == False and if there is an import error, the
      function handles the error and returns None.

    Returns
    -------

    The corresponding FFT class.

    """
    if method.startswith('fft2d.') or method.startswith('fft3d.'):
        method = 'fluidfft.' + method
    elif method.startswith('fluidfft.fft2d.'):
        pass
    else:
        raise ValueError

    try:
        mod = _import_module(method)
    except ImportError:
        if raise_import_error:
            raise ImportError(method)
        else:
            print('ImportError:', method)
            return None

    return mod.FFTclass


def create_fft_object(method, n0, n1, n2=None):
    """Helper for creating fft objects.

    Parameters
    ----------

    method : str
      Name of module or string characterizing a method. It has to correspond to
      a module of fluidfft. The first part "fluidfft." of the module "path" can
      be omitted.

    n0, n1, n2 : int
      Dimensions of the real space array (in sequential).

    Returns
    -------

    The corresponding FFT object.


    """

    if n2 is None:
        if method.startswith('fft2d.'):
            method = 'fluidfft.' + method
        elif method.startswith('fluidfft.fft2d.'):
            pass
        elif (method.startswith('fft3d.') and
              method.startswith('fluidfft.fft3d.')):
            raise ValueError('Arguments incompatible')
        else:
            method = 'fluidfft.fft2d.' + method
    else:
        if method.startswith('fft3d.'):
            method = 'fluidfft.' + method
        elif method.startswith('fluidfft.fft3d.'):
            pass
        elif (method.startswith('fft2d.') and
              method.startswith('fluidfft.fft2d.')):
            raise ValueError('Arguments incompatible')
        else:
            method = 'fluidfft.fft3d.' + method

    try:
        mod = _import_module(method)
    except ImportError:
        raise ImportError(method)

    if n2 is None:
        return mod.FFTclass(n0, n1)
    else:
        return mod.FFTclass(n0, n1, n2)
